from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *

from xicam.gui.static import path


class SearchLineEdit(QLineEdit):
    def __init__(self, text='', clearable = True, parent=None):
        QLineEdit.__init__(self, text=text, parent=parent)

        searchPixmap = QPixmap(str(path("icons/search.png")))

        clearPixmap = QPixmap(str(path("icons/clear.png")))
        self.clearButton = QToolButton(self)
        self.clearButton.setIcon(QIcon(clearPixmap))
        self.clearButton.setIconSize(QSize(16, 16))
        self.clearButton.setCursor(Qt.ArrowCursor)
        self.clearButton.setStyleSheet("QToolButton { border: none; padding: 0 px;}")
        self.clearButton.hide()

        if clearable:
            self.clearButton.clicked.connect(self.clear)
            self.textChanged.connect(self.updateCloseButton)

        self.searchButton = QToolButton(self)
        self.searchButton.setIcon(QIcon(searchPixmap))
        self.searchButton.setIconSize(QSize(16, 16))
        self.searchButton.setStyleSheet("QToolButton { border: none; padding: 0 px;}")

        frameWidth = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        self.setStyleSheet("QLineEdit { padding-left: %spx; padding - right: % spx;} " %
                           (self.searchButton.sizeHint().width() + frameWidth + 1,
                            self.clearButton.sizeHint().width() + frameWidth + 1))
        msz = self.minimumSizeHint()
        self.setMinimumSize(max(msz.width(),
                                self.searchButton.sizeHint().width() +
                                self.clearButton.sizeHint().width() +
                                frameWidth * 2 + 2),
                            max(msz.height(), self.clearButton.sizeHint().height() + frameWidth * 2 + 2))

    #        self.searchMenu = QtGui.QMenu(self.searchButton)
    #        self.searchButton.setMenu(self.searchMenu)
    #        self.searchMenu.addAction("Google")
    #        self.searchButton.setPopupMode(QtGui.QToolButton.InstantPopup)


    def resizeEvent(self, event):
        sz = self.clearButton.sizeHint()
        frameWidth = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        self.clearButton.move(self.rect().right() - frameWidth -
                              sz.width(),
                              (self.rect().bottom() + 1 - sz.height()) / 2)
        self.searchButton.move(self.rect().left() + 1,
                               (self.rect().bottom() + 1 - sz.height()) / 2)

    def updateCloseButton(self, text):
        if text:
            self.clearButton.setVisible(True)
        else:
            self.clearButton.setVisible(False)



import requests
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from xicam.gui.static import path
from xicam.gui.widgets.searchlineedit import SearchLineEdit

from xicam.plugins import SettingsPlugin
from xicam.plugins import cammart


class CamMartSettingsPlugin(SettingsPlugin):
    """
    A built-in settings plugin to configure installed packages
    """
    name = 'Plugins'

    def __init__(self):
        # Setup UI
        self.widget = QWidget()
        self.widget.setLayout(QHBoxLayout())
        self.listview = QListView()
        self.packagesmodel = QStandardItemModel()
        self.listview.setModel(self.packagesmodel)

        self.plugintoolbar = QToolBar()
        self.plugintoolbar.setOrientation(Qt.Vertical)
        self.plugintoolbar.addAction(QIcon(str(path('icons/plus.png'))),
                                     'Add plugin',
                                     self.addplugin)
        self.plugintoolbar.addAction(QIcon(str(path('icons/minus.png'))),
                                     'Remove plugin',
                                     self.removeplugin)
        self.widget.layout().addWidget(self.listview)
        self.widget.layout().addWidget(self.plugintoolbar)
        super(CamMartSettingsPlugin, self).__init__(QIcon(str(path('icons/box.png'))),
                                                    self.name,
                                                    self.widget)

        # Refresh packages list
        self.refresh()

    def refresh(self):
        """
        Refresh the list of packages by updating the model
        """
        self.packagesmodel.clear()
        for name, scheme in cammart.pkg_registry.items():
            self.packagesmodel.appendRow(QStandardItem(name))

    def addplugin(self):
        """
        Open the CamMart install dialog
        """
        self._dialog = CamMartInstallDialog()
        self._dialog.exec_()
        self.refresh()

    def removeplugin(self):
        """
        Uninstalls a plugin
        """
        if self.listview.selectedIndexes():
            cammart.uninstall(self.packagesmodel.itemFromIndex(self.listview.selectedIndexes()[0]).text())
            self.refresh()

    def save(self):  # This class has no settings to save, it is driven by the packages list
        return None  # self.parameter.saveState()

    def restore(self, state):
        pass  # self.parameter.restoreState(state)


repositories = ['cam.lbl.gov:5000']


class CamMartInstallDialog(QDialog):
    def __init__(self):
        super(CamMartInstallDialog, self).__init__()

        # Set size and position
        self.setGeometry(0, 0, 800, 500)
        frameGm = self.frameGeometry()
        screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
        centerPoint = QApplication.desktop().screenGeometry(screen).center()
        frameGm.moveCenter(centerPoint)
        self.move(frameGm.topLeft())

        # Setup ListView
        self.packagesWidget = QListView()
        self.packagesWidget.setMovement(QListView.Static)
        self.packagesWidget.setSelectionMode(QAbstractItemView.SingleSelection)

        # Setup Model
        self.packagesModel = QStandardItemModel()
        self.packagesWidget.setModel(self.packagesModel)
        self.packageInfoWidget = PackageInfoWidget(self.packagesWidget)
        self.packagesWidget.selectionModel().currentChanged.connect(self.packageInfoWidget.refresh)

        # Setup dialog buttons
        self.installButton = QPushButton("&Install Package")
        self.manageButton = QPushButton("&Manage Repositories")
        self.installButton.clicked.connect(self.install)
        self.manageButton.clicked.connect(self.manage)
        self.buttonboxWidget = QDialogButtonBox()
        self.buttonboxWidget.addButton(self.installButton, QDialogButtonBox.ActionRole)
        self.buttonboxWidget.addButton(self.manageButton, QDialogButtonBox.ActionRole)

        # Setup splitter layout
        splitter = QSplitter()
        splitter.addWidget(self.packagesWidget)
        splitter.addWidget(self.packageInfoWidget)

        # Setup search box
        self.searchbox = SearchLineEdit()

        # Compose main layout
        mainLayout = QVBoxLayout()
        mainLayout.addWidget(self.searchbox)
        mainLayout.addWidget(splitter)
        mainLayout.addSpacing(12)
        mainLayout.addWidget(self.buttonboxWidget)
        self.setLayout(mainLayout)
        self.setWindowTitle("Install Packages...")

        # Set modality
        self.setModal(True)

        # Load packages into view
        self.refresh()

    def refresh(self):  # TODO: display dialog when query fails (no connection)
        # Clear model
        self.packagesModel.clear()

        # For each repo
        for repo in repositories:
            # TODO: check behavior for >25 items (pagesize)
            # For each package
            for packageinfo in eval(requests.get(f'http://{repo}/pluginpackages').content)["_items"]:
                # Add an item to the model
                item = QStandardItem(packageinfo['name'])
                item.info = packageinfo
                self.packagesModel.appendRow(item)

    def install(self):
        # Install the selected package using cammart
        cammart.install(self.packagesModel.itemFromIndex(self.packagesWidget.selectedIndexes()[0]).text())

    def manage(self):
        pass


class PackageInfoWidget(QTextEdit):
    def __init__(self, view):
        super(PackageInfoWidget, self).__init__()

        self.view = view

    def refresh(self, current, previous):
        # Get info of current package from the model's item
        info = self.view.model().itemFromIndex(current).info

        # Display info
        self.setText(
            f"""
            <h1>{info['name']}</h1>
            <p>{info['documentation'].get('description')}</p>
            <p>Version: {info['documentation'].get('version')}</p>
            <p>Authors: {", ".join(info['documentation'].get('authors'))}</p>
            <p>Publication: {info['documentation'].get('publication')}</p>
            <p>Reference: {info['documentation'].get('reference')}</p>
            <p>Keywords: {", ".join(info['documentation'].get('keywords'))}</p>
            """)



# Adapted from http://doc.qt.io/qt-5/qtwidgets-dialogs-configdialog-configdialog-cpp.html under BSD


# TODO QtModern, QtDarkStyle
# TODO Add remotes config
# TODO Add usage statistics config
# TODO QSettings

from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *

from xicam.plugins import manager as pluginmanager


class ConfigDialog(QDialog):
    def __init__(self):
        super(ConfigDialog, self).__init__()

        # Set size and position
        self.setGeometry(0, 0, 900, 550)
        frameGm = self.frameGeometry()
        screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
        centerPoint = QApplication.desktop().screenGeometry(screen).center()
        frameGm.moveCenter(centerPoint)
        self.move(frameGm.topLeft())

        self.contentsWidget = QListView()
        self.contentsWidget.setViewMode(QListView.IconMode)
        # self.contentsWidget.setIconSize(QSize(96, 84))
        self.contentsWidget.setMovement(QListView.Static)
        self.contentsWidget.setMaximumWidth(174)
        self.contentsWidget.setSpacing(12)
        self.contentsWidget.setSelectionMode(QAbstractItemView.SingleSelection)

        self.contentsModel = QStandardItemModel()
        self.contentsWidget.setModel(self.contentsModel)
        self.contentsWidget.selectionModel().currentChanged.connect(self.changePage)

        self.buttonboxWidget = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel | QDialogButtonBox.Apply | QDialogButtonBox.Help)
        self.buttonboxWidget.button(QDialogButtonBox.Ok).clicked.connect(self.ok)
        self.buttonboxWidget.button(QDialogButtonBox.Apply).clicked.connect(self.apply)
        self.buttonboxWidget.button(QDialogButtonBox.Cancel).clicked.connect(self.close)

        self.pagesWidget = QStackedWidget()

        horizontalLayout = QHBoxLayout()
        horizontalLayout.addWidget(self.contentsWidget)
        horizontalLayout.addWidget(self.pagesWidget, 1)

        mainLayout = QVBoxLayout()
        mainLayout.addLayout(horizontalLayout)
        # mainLayout.addStretch(1)
        mainLayout.addSpacing(12)
        mainLayout.addWidget(self.buttonboxWidget)

        self.setLayout(mainLayout)
        self.setWindowTitle("Config Dialog")

        # Set modality
        self.setModal(True)

        self.createIcons()
        self.restore()

    def createIcons(self):
        for pluginInfo in pluginmanager.getPluginsOfCategory('SettingsPlugin'):
            item = QStandardItem(pluginInfo.plugin_object.icon, pluginInfo.plugin_object.name)
            item.widget = pluginInfo.plugin_object.widget
            item.setTextAlignment(Qt.AlignHCenter)
            item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
            item.setSizeHint(QSize(136, 80))
            self.contentsModel.appendRow(item)

    def changePage(self, current, previous):
        if not current:
            current = previous
        current = self.contentsModel.itemFromIndex(current)
        self.pagesWidget.addWidget(current.widget)
        self.pagesWidget.setCurrentWidget(current.widget)

    def pluginsChanged(self):
        self.createIcons()

    def restore(self):
        try:
            for pluginInfo in pluginmanager.getPluginsOfCategory('SettingsPlugin'):
                pluginInfo.plugin_object.restore(QSettings().value(pluginInfo.name))
        except (AttributeError, TypeError, SystemError):
            # No settings saved
            pass
        self.apply()

    def ok(self):
        self._empty()
        self.apply()
        self.accept()

    def apply(self):
        for pluginInfo in pluginmanager.getPluginsOfCategory('SettingsPlugin'):
            QSettings().setValue(pluginInfo.name, pluginInfo.plugin_object.save())

    def close(self):
        self._empty()
        self.restore()
        self.reject()

    def _empty(self):
        """
        Disown all widget children (otherwise their c++ objects are force deleted when the dialog closes).
        Must be run in reverse to avoid index update errors
        """
        for i in reversed(range(self.pagesWidget.count())):
            self.pagesWidget.widget(i).setParent(None)

    def closeEvent(self, event):
        self.close()
        event.accept()

    def keyPressEvent(self, e: QKeyEvent):
        if e.key() != Qt.Key_Escape:
            super(ConfigDialog, self).keyPressEvent(e)
        else:
            self.close()



import time
from functools import partial
from xicam.core import msg

from qtpy.QtCore import *


class ThreadManager(QObject):
    sigStateChanged = Signal()

    def __init__(self):
        super(ThreadManager, self).__init__()
        self._threads = []

    @property
    def threads(self):
        self.purge()
        return self._threads

    def purge(self):
        self._threads = [thread for thread in self._threads if not thread.purge]
        for thread in self._threads:
            if thread.done or thread.cancelled or thread.exception:
                thread.purge = True

    def append(self, thread):
        self._threads.append(thread)


manager = ThreadManager()


class QThreadFuture(QObject):
    sigCallback = Signal()
    sigFinished = Signal()
    sigExcept = Signal(Exception)

    def __init__(self, method, *args, callback_slot=None, finished_slot=None, except_slot=None, default_exhandle=True,
                 lock=None, threadkey: str = None, showBusy=True, **kwargs):
        super(QThreadFuture, self).__init__()

        # Auto-Kill other threads with same threadkey
        if threadkey:
            for thread in manager.threads:
                if thread.threadkey == threadkey:
                    thread.cancel()
        self.threadkey = threadkey

        self.callback_slot = callback_slot
        # if callback_slot: self.sigCallback.connect(callback_slot)
        if finished_slot: self.sigFinished.connect(finished_slot)
        if except_slot: self.sigExcept.connect(except_slot)
        self.method = method
        self.args = args
        self.kwargs = kwargs

        self.cancelled = False
        self.running = False
        self.done = False
        self.exception = None
        self.purge = False
        self.thread = None
        self.showBusy = showBusy

        manager.append(self)

    def start(self):
        if not self.thread:
            self.thread = QThread()
            self.thread.run = partial(self.run, *self.args, **self.kwargs)
            self.thread.start()
        else:
            raise ValueError('Thread could not be started; it is already running.')

    def run(self, *args, **kwargs):
        self.cancelled = False
        self.running = True
        self.done = False
        self.exception = None
        if self.showBusy: invoke_in_main_thread(msg.showBusy)
        try:
            for self._result in self._run(*args, **kwargs):
                if not isinstance(self._result, tuple): self._result = (self._result,)
                if self.callback_slot: invoke_in_main_thread(self.callback_slot, *self._result)
                self.running = False

        except Exception as ex:
            self.exception = ex
            self.sigExcept.emit(ex)
            msg.logMessage(f'Error in thread: '
                           f'Method: {self.method.__name__}\n'
                           f'Args: {self.args}\n'
                           f'Kwargs: {self.kwargs}', level=msg.ERROR)
            msg.logError(ex)
        else:
            self.done = True
            self.sigFinished.emit()
        finally:
            invoke_in_main_thread(msg.showReady)

    def _run(self, *args, **kwargs):
        yield self.method(*self.args, **self.kwargs)

    def result(self):
        while not self.done and not self.exception:
            time.sleep(.1)
        if self.exception: return self.exception
        return self._result

    def cancel(self):
        self.thread.quit()
        self.thread.wait()
        self.cancelled = True


class QThreadFutureIterator(QThreadFuture):
    def _run(self, *args, **kwargs):
        yield from self.method(*self.args, **self.kwargs)


class InvokeEvent(QEvent):
    EVENT_TYPE = QEvent.Type(QEvent.registerEventType())

    def __init__(self, fn, *args, **kwargs):
        QEvent.__init__(self, InvokeEvent.EVENT_TYPE)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs


class Invoker(QObject):
    def event(self, event):
        try:
            if hasattr(event.fn, 'signal'):  # check if invoking a signal or a callable
                event.fn.emit(*event.args, *event.kwargs.values())
            else:
                event.fn(*event.args, **event.kwargs)
            return True
        except Exception as ex:
            msg.logMessage('QThreadFuture callback could not be invoked.', level=msg.ERROR)
            msg.logError(ex)
        return False


_invoker = Invoker()


def invoke_in_main_thread(fn, *args, **kwargs):
    # print 'attempt invoke:',fn,args,kwargs
    QCoreApplication.postEvent(_invoker,
                               InvokeEvent(fn, *args, **kwargs))



from qtpy.QtWidgets import *
from qtpy.QtGui import *
from qtpy.QtCore import *
from typing import List
from functools import partial


class TabView(QTabWidget):
    def __init__(self, model=None, widgetcls=None, field=None, **kwargs):
        super(TabView, self).__init__()
        self.setTabBar(ContextMenuTabBar())
        self.kwargs = kwargs

        self.setWidgetClass(widgetcls)
        self.model = None
        self.selectionmodel = None  # type: TabItemSelectionModel
        if model: self.setModel(model)
        self.field = field

    def setModel(self, model: QStandardItemModel):
        self.model = model
        self.selectionmodel = TabItemSelectionModel(self)
        model.dataChanged.connect(self.dataChanged)
        self.tabCloseRequested.connect(self.closeTab)

        self.setTabsClosable(True)
        self.setDocumentMode(True)

    def dataChanged(self, start, end):
        for i in range(self.model.rowCount()):

            if self.widget(i):
                if self.widget(i).header == self.model.item(i).header:
                    continue
            self.setCurrentIndex(
                self.insertTab(i, self.widgetcls(self.model.item(i).header, self.field, **self.kwargs),
                               self.model.item(i).text()))

        for i in reversed(range(self.model.rowCount(), self.count())):
            self.removeTab(i)

    def setWidgetClass(self, cls):
        self.widgetcls = cls

    def currentHeader(self):
        return self.model.item(self.currentIndex())

    def closeTab(self, i):
        self.removeTab(i)
        self.model.removeRow(i)


class TabViewSynchronizer(QObject):
    def __init__(self, tabviews: List[TabView]):
        super(TabViewSynchronizer, self).__init__()
        self.tabviews = tabviews
        for tabview in tabviews:
            tabview.currentChanged.connect(partial(self.sync, sourcetabview=tabview))
            tabview.tabCloseRequested.connect(partial(self.sync, sourcetabview=tabview))

    def sync(self, index, sourcetabview):
        for tabview in self.tabviews:
            if tabview is sourcetabview: continue
            tabview.setCurrentIndex(index)
            tabview.dataChanged(None, None)


class ContextMenuTabBar(QTabBar):
    def __init__(self):
        super(ContextMenuTabBar, self).__init__()
        self.contextMenu = QMenu()
        self.closeaction = QAction('&Close')
        self.closeaction.triggered.connect(self.close)
        self.closeothersaction = QAction('Close &Others')
        self.closeothersaction.triggered.connect(self.closeothers)
        self.closeallaction = QAction('Close &All')
        self.closeallaction.triggered.connect(self.closeall)
        self.contextMenu.addActions([self.closeaction, self.closeothersaction, self.closeallaction])
        self._rightclickedtab = None

    def close(self):
        self.tabCloseRequested.emit(self._rightclickedtab)

    def closeothers(self):
        for i in reversed(range(self.count())):
            if i != self._rightclickedtab:
                self.tabCloseRequested.emit(i)

    def closeall(self):
        for i in reversed(range(self.count())):
            self.tabCloseRequested.emit(i)

    def mousePressEvent(self, event: QMouseEvent):
        super(ContextMenuTabBar, self).mousePressEvent(event)
        self._rightclickedtab = self.tabAt(event.pos())
        if self._rightclickedtab != -1:
            if event.button() == Qt.RightButton:
                self.contextMenu.popup(self.mapToGlobal(event.pos()))


class TabItemSelectionModel(QItemSelectionModel):
    def __init__(self, tabview: TabView):
        super(TabItemSelectionModel, self).__init__(tabview.model)
        self.tabview = tabview

    def currentIndex(self):
        return self.tabview.currentIndex()



from pyqtgraph.flowchart import FlowchartCtrlWidget, Flowchart, Node, NodeLibrary
from xicam.core.execution.workflow import Workflow
from xicam.plugins import manager as pluginmanager


# LIBRARY = NodeLibrary.NodeLibrary()
# for plugin in pluginmanager.getPluginsOfCategory('ProcessingPlugin'):
#     nodecls = type(plugin.name,(Node,))
#     LIBRARY.addNodeType(plugin.name,plugin.name)


class DaskFlow(FlowchartCtrlWidget):

    def __init__(self):
        self.flowchart = Flowchart()
        super(DaskFlow, self).__init__(self.flowchart)

    def fromDask(self, workflow: Workflow):
        for process in workflow.processes:
            node = Node(process.name, terminals={'inputTerminalName': {'io': 'in'},
                                                 'outputTerminalName': {'io': 'out'}})
            self.flowchart.addNode(node, process.name)


# -*- coding: utf-8 -*-
"""
This example demonstrates a very basic use of flowcharts: filter data,
displaying both the input and output of the filter. The behavior of
the filter can be reprogrammed by the user.

Basic steps are:
  - create a flowchart and two plots
  - input noisy data to the flowchart
  - flowchart connects data to the first plot, where it is displayed
  - add a gaussian filter to lowpass the data, then display it in the second plot.
"""
# import initExample ## Add path to library (just for examples; you do not need this)
#
#
# from pyqtgraph.flowchart import Flowchart
# from pyqtgraph.Qt import QtGui, QtCore
# import pyqtgraph as pg
# import numpy as np
# import pyqtgraph.metaarray as metaarray
#
# app = QtGui.QApplication([])
#
# ## Create main window with grid layout
# win = QtGui.QMainWindow()
# win.setWindowTitle('pyqtgraph example: Flowchart')
# cw = QtGui.QWidget()
# win.setCentralWidget(cw)
# layout = QtGui.QGridLayout()
# cw.setLayout(layout)
#
# ## Create flowchart, define input/output terminals
# fc = Flowchart(terminals={
#     'dataIn': {'io': 'in'},
#     'dataOut': {'io': 'out'}
# })
# w = fc.widget()
#
# ## Add flowchart control panel to the main window
# layout.addWidget(fc.widget(), 0, 0, 2, 1)
#
# ## Add two plot widgets
# pw1 = pg.PlotWidget()
# pw2 = pg.PlotWidget()
# layout.addWidget(pw1, 0, 1)
# layout.addWidget(pw2, 1, 1)
#
# win.show()
#
# ## generate signal data to pass through the flowchart
# data = np.random.normal(size=1000)
# data[200:300] += 1
# data += np.sin(np.linspace(0, 100, 1000))
# data = metaarray.MetaArray(data, info=[{'name': 'Time', 'values': np.linspace(0, 1.0, len(data))}, {}])
#
# ## Feed data into the input terminal of the flowchart
# fc.setInput(dataIn=data)
#
# ## populate the flowchart with a basic set of processing nodes.
# ## (usually we let the user do this)
# plotList = {'Top Plot': pw1, 'Bottom Plot': pw2}
#
# pw1Node = fc.createNode('PlotWidget', pos=(0, -150))
# pw1Node.setPlotList(plotList)
# pw1Node.setPlot(pw1)
#
# pw2Node = fc.createNode('PlotWidget', pos=(150, -150))
# pw2Node.setPlot(pw2)
# pw2Node.setPlotList(plotList)
#
# fNode = fc.createNode('GaussianFilter', pos=(0, 0))
# fNode.ctrls['sigma'].setValue(5)
# fc.connectTerminals(fc['dataIn'], fNode['In'])
# fc.connectTerminals(fc['dataIn'], pw1Node['In'])
# fc.connectTerminals(fNode['Out'], pw2Node['In'])
# fc.connectTerminals(fNode['Out'], fc['dataOut'])
#
#
#
# ## Start Qt event loop unless running in interactive mode or using pyside.
# if __name__ == '__main__':
#     import sys
#     if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
#         QtGui.QApplication.instance().exec_()



from qtpy.QtWidgets import *
from qtpy.QtCore import *
from qtpy.QtGui import *
from ..clientonlymodels.LocalFileSystemResource import LocalFileSystemResourcePlugin
from xicam.gui.static import path
from xicam.core.data import NonDBHeader
from xicam.plugins import manager as pluginmanager
from xicam.plugins.DataResourcePlugin import DataSourceListModel
from .searchlineedit import SearchLineEdit
from urllib import parse
from pathlib import Path
from functools import partial
import os, webbrowser
from xicam.gui.widgets.tabview import ContextMenuTabBar




class BrowserTabWidget(QTabWidget):
    def __init__(self, parent=None):
        super(BrowserTabWidget, self).__init__(parent)
        self.setContentsMargins(0,0,0,0)

class DataBrowser(QWidget):
    sigOpen = Signal(NonDBHeader)
    sigPreview = Signal(NonDBHeader)

    def __init__(self, browserview):
        super(DataBrowser, self).__init__()

        hbox = QHBoxLayout()
        vbox = QVBoxLayout()
        vbox.setContentsMargins(0,0,0,0)
        vbox.setSpacing(0)
        hbox.setContentsMargins(0,0,0,0)
        hbox.setSpacing(0)
        self.setContentsMargins(0,0,0,0)

        self.browserview = browserview
        self.browserview.sigOpen.connect(self.sigOpen)
        self.browserview.sigPreview.connect(self.sigPreview)
        self.browserview.sigOpenExternally.connect(self.openExternally)
        self.browserview.sigConfigChanged.connect(self.pushConfigtoURI)
        self.toolbar = QToolBar()
        self.toolbar.addAction(QIcon(QPixmap(str(path('icons/up.png')))), 'Move up directory', self.moveUp)
        # self.toolbar.addAction(QIcon(QPixmap(str(path('icons/filter.png')))), 'Filter')
        self.toolbar.addAction(QIcon(QPixmap(str(path('icons/refresh.png')))), 'Refresh', self.hardRefreshURI)
        self.toolbar.setToolButtonStyle(Qt.ToolButtonIconOnly)
        self.URILineEdit = SearchLineEdit('',clearable=False)
        self.pushConfigtoURI()

        hbox.addWidget(self.toolbar)
        hbox.addWidget(self.URILineEdit)
        vbox.addLayout(hbox)
        vbox.addWidget(self.browserview)
        self.setLayout(vbox)

        self.URILineEdit.textChanged.connect(self.softRefreshURI)
        self.URILineEdit.returnPressed.connect(self.softRefreshURI)  # hard refresh
        self.URILineEdit.focusOutEvent = self.softRefreshURI  # hard refresh

        self.hardRefreshURI()

    def pushURItoConfig(self):
        uri = parse.urlparse(self.URILineEdit.text())
        config = self.browserview.model.config
        config['scheme'] = uri.scheme
        config['host'] = uri.netloc
        config['path'] = uri.path
        config['query'] = uri.query
        config['fragment'] = uri.fragment
        config['params'] = uri.params
        config['user'] = uri.username
        config['port'] = uri.port
        config['password'] = uri.password
        print('config:', config)
        return config

    def pushConfigtoURI(self):
        config = self.browserview.model.config
        uri = parse.ParseResult(scheme=config.get('scheme',''),
                                netloc=config.get('host',''),
                                path=config.get('path',''),
                                params=config.get('params',''),
                                query=config.get('query',''),
                                fragment=config.get('fragment',''))
        uri = parse.urlunparse(uri)
        self.URILineEdit.setText(uri)
        return config


    def hardRefreshURI(self, *_, **__):
        self.pushURItoConfig()
        self.browserview.refresh()

    def moveUp(self):
        config = self.pushURItoConfig()
        config['path'] = str(Path(config['path']).parent)
        self.browserview.refresh()
        self.pushConfigtoURI()

    def openExternally(self, uri):
        webbrowser.open(uri)

    softRefreshURI = hardRefreshURI


class BrowserTabBar(ContextMenuTabBar):
    sigAddBrowser = Signal(DataBrowser, str)

    def __init__(self, tabwidget: QTabWidget):
        super(BrowserTabBar, self).__init__()

        self.tabwidget = tabwidget
        self.tabwidget.setTabBar(self)

        self.setExpanding(False)
        self.setTabsClosable(True)

        plusPixmap = QPixmap(str(path("icons/plus.png")))
        self.plusIcon = QIcon(plusPixmap)
        # self.plus_button.setToolTip('Open a new browser')
        # self.plus_button.setParent(self)
        # self.plus_button.setMaximumSize(32, 32)
        # self.plus_button.setMinimumSize(32, 32)
        # self.plus_button.clicked.connect(self.sigAddBrowser.emit)
        tab = self.addTab(self.plusIcon, '')
        try:
            self.tabButton(tab, QTabBar.RightSide).resize(0, 0)
            self.tabButton(tab, QTabBar.RightSide).hide()
        except AttributeError:
            self.tabButton(tab, QTabBar.LeftSide).resize(0, 0)
            self.tabButton(tab, QTabBar.LeftSide).hide()
        # self.movePlusButton()  # Move to the correct location
        # self.setDocumentMode(True)
        self.currentChanged.connect(self.tabwidget.setCurrentIndex)
        self.installEventFilter(self)

    def addTab(self, *args, **kwargs):
        return self.insertTab(self.count() - 1, *args, **kwargs)

    def eventFilter(self, object, event):
        try:
            if object == self and event.type() in [QEvent.MouseButtonPress,
                                                   QEvent.MouseButtonRelease] and event.button() == Qt.LeftButton:

                if event.type() == QEvent.MouseButtonPress:
                    tabIndex = object.tabAt(event.pos())
                    if tabIndex == self.count() - 1:
                        self.showMenu(self.mapToGlobal(event.pos()))
                        return True
            return False
        except Exception as e:
            print("Exception raised in eventfilter", e)

    def showMenu(self, pos):
        self.menu = QMenu()
        # Add all resource plugins
        self.actions = {}
        for plugin in pluginmanager.getPluginsOfCategory('DataResourcePlugin'):
            self.actions[plugin.name] = QAction(plugin.name)
            self.actions[plugin.name].triggered.connect(partial(self._addBrowser, plugin))
            self.menu.addAction(self.actions[plugin.name])

        self.menu.popup(pos)

    def _addBrowser(self, plugin):
        self.sigAddBrowser.emit(DataBrowser(DataResourceList(DataSourceListModel(plugin.plugin_object))), plugin.name)


class DataResourceView(QObject):
    def __init__(self, model):
        super(DataResourceView, self).__init__()
        self.model = model
        self.setModel(self.model)
        self.doubleClicked.connect(self.open)
        self.setSelectionMode(self.ExtendedSelection)
        self.setSelectionBehavior(self.SelectRows)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.menuRequested)

        self.menu = QMenu()
        standardActions = [QAction('Open', self),
                           QAction('Open Externally', self),
                           QAction('Enable/Disable Streaming', self),
                           QAction('Delete', self)]
        self.menu.addActions(standardActions)
        standardActions[0].triggered.connect(self.open)
        standardActions[1].triggered.connect(self.openExternally)

    def menuRequested(self, position):
        self.menu.exec_(self.viewport().mapToGlobal(position))

    def open(self, index):
        pass

    def currentChanged(self, current, previous):
        pass

    def openExternally(self, uri:str):
        pass


class DataResourceTree(QTreeView, DataResourceView):
    sigOpen = Signal(NonDBHeader)
    sigOpenPath = Signal(str)
    sigOpenExternally = Signal(str)
    sigPreview = Signal(NonDBHeader)
    sigConfigChanged = Signal()

    def __init__(self, *args):
        super(DataResourceTree, self).__init__(*args)

    def refresh(self):
        self.model.refresh()
        self.setRootIndex(self.model.index(self.model.path))


class DataResourceList(QListView, DataResourceView):
    sigOpen = Signal(NonDBHeader)
    sigOpenPath = Signal(str)
    sigOpenExternally = Signal(str)
    sigPreview = Signal(NonDBHeader)
    sigConfigChanged = Signal()

    def refresh(self):
        self.model.refresh()


class LocalFileSystemTree(DataResourceTree):

    def __init__(self):
        self.model = LocalFileSystemResourcePlugin()
        super(LocalFileSystemTree, self).__init__(self.model)

    def open(self, _):
        indexes = self.selectionModel().selectedRows()
        if len(indexes)==1:
            path = self.model.filePath(indexes[0])
            if os.path.isdir(path):
                self.model.path = path
                self.setRootIndex(indexes[0])
                self.sigConfigChanged.emit()
                return
        self.sigOpen.emit(self.model.getHeader(indexes))

    def currentChanged(self, current, previous):
        if current.isValid():
            self.sigPreview.emit(self.model.getHeader([current]))



    def openExternally(self, uri:str):
        indexes = self.selectionModel().selectedRows()
        for index in indexes:
            self.sigOpenExternally.emit(self.model.filePath(index))


class DataResourceBrowser(QWidget):
    sigOpen = Signal(NonDBHeader)
    sigPreview = Signal(NonDBHeader)

    def __init__(self):
        super(DataResourceBrowser, self).__init__()
        vbox = QVBoxLayout()
        vbox.setSpacing(0)
        vbox.setContentsMargins(0,0,0,0)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setMinimumSize(QSize(250, 400))

        self.browsertabwidget = BrowserTabWidget(self)
        self.browsertabbar = BrowserTabBar(self.browsertabwidget)
        self.browsertabbar.sigAddBrowser.connect(self.addBrowser)
        self.browsertabbar.tabCloseRequested.connect(self.closetab)

        # Add the required 'Local' browser
        self.addBrowser(DataBrowser(LocalFileSystemTree()), 'Local', closable=False)
        self.browsertabbar.setCurrentIndex(0)

        vbox.addWidget(self.browsertabwidget)

        self.setLayout(vbox)

        self.sigOpen.connect(print)

    def closetab(self, i):
        if hasattr(self.browsertabwidget.widget(i), 'closable'):
            if self.browsertabwidget.widget(i).closable:
                self.browsertabwidget.removeTab(i)

    def sizeHint(self):
        return QSize(250, 400)

    def addBrowser(self, databrowser:DataBrowser, text:str, closable:bool=True):
        databrowser.sigOpen.connect(self.sigOpen)
        databrowser.sigPreview.connect(self.sigPreview)
        databrowser.closable = closable
        tab = self.browsertabwidget.addTab(databrowser, text)
        # self.browsertabbar.addTab(text)
        if closable is False:
            try:
                self.browsertabbar.tabButton(tab, QTabBar.RightSide).resize(0, 0)
                self.browsertabbar.tabButton(tab, QTabBar.RightSide).hide()
            except AttributeError:
                self.browsertabbar.tabButton(tab, QTabBar.LeftSide).resize(0, 0)
                self.browsertabbar.tabButton(tab, QTabBar.LeftSide).hide()


from collections import OrderedDict

import qdarkstyle
from qtmodern import styles
from qtpy.QtGui import *
from qtpy.QtWidgets import QApplication
from xicam.gui.static import path
import pyqtgraph as pg
from xicam import plugins

from xicam.plugins import SettingsPlugin


def setDefault():
    QApplication.instance().setStyleSheet("")


def setDark():
    QApplication.instance().setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())


def setModern():
    styles.dark(QApplication.instance())


def setUglyGreen():
    QApplication.instance().setStyleSheet("QWidget {background-color: darkgreen;}")


def setPlotWhite():
    pg.setConfigOption('background', 'w')
    pg.setConfigOption('foreground', 'k')


def setPlotDefault():
    pass


if plugins.qt_is_safe:
    AppearanceSettingsPlugin = SettingsPlugin.fromParameter(QIcon(str(path('icons/colors.png'))),
                                                        'Appearance',
                                                        [dict(name='Theme',
                                                              values=OrderedDict([('Default', setDefault),
                                                                                  ('QDarkStyle', setDark),
                                                                                  ('QtModern', setModern),
                                                                                  ('UglyGreen', setUglyGreen)]),
                                                              type='list'),
                                                         dict(name='Plot Theme (requires restart)',
                                                              values=OrderedDict([('Default (Dark)', setPlotDefault),
                                                                                  ('Publication (White)',
                                                                                   setPlotWhite)]),
                                                              type='list')]
                                                        )


    def apply(self):
        self.parameter['Theme']()
        self.parameter['Plot Theme (requires restart)']()


    AppearanceSettingsPlugin.apply = apply



from typing import Callable
import sys

from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *

from xicam.gui import static


class XicamSplashScreen(QSplashScreen):
    minsplashtime = 3000

    def __init__(self, mainwindow: Callable[[], QMainWindow] = None,
                 f: int = Qt.WindowStaysOnTopHint | Qt.SplashScreen):
        """
        A QSplashScreen customized to display an animated gif. The splash triggers launch when clicked.

        After minsplashtime, this splash waits until the animation finishes before triggering the launch.

        Parameters
        ----------
        mainwindow  :   class
            Subclass of QMainWindow to display after splashing
        f           :   int
            Extra flags (see base class)
        """

        # Get logo movie from relative path
        self.movie = QMovie(str(static.path('images/animated_logo.gif')))

        # Setup drawing
        self.movie.frameChanged.connect(self.paintFrame)
        self.movie.jumpToFrame(1)
        self.pixmap = QPixmap(self.movie.frameRect().size())
        super(XicamSplashScreen, self).__init__(self.pixmap, f)
        self.setMask(self.pixmap.mask())
        self.movie.finished.connect(self.restartmovie)

        self._launching = False
        self._launchready = False
        self.timer = QTimer(self)
        self.mainwindow = mainwindow

        # Start splashing
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.show()
        self.raise_()
        self.activateWindow()
        QApplication.instance().setActiveWindow(self)
        if '--quicksplash' in sys.argv:
            self.execlaunch()
        else:
            # Setup timed triggers for launching the QMainWindow
            self.timer.singleShot(self.minsplashtime, self.launchwindow)

    def showMessage(self, message: str, color=Qt.white):
        # TODO: Make this work.
        super(XicamSplashScreen, self).showMessage(message, color)

    def mousePressEvent(self, *args, **kwargs):
        # TODO: Apparently this doesn't work?
        self.timer.stop()
        self.execlaunch()

    def showEvent(self, event):
        """
        Start the animation when shown
        """
        self.movie.start()

    def paintFrame(self):
        """
        Paint the current frame
        """
        self.pixmap = self.movie.currentPixmap()
        self.setMask(self.pixmap.mask())
        self.setPixmap(self.pixmap)
        self.movie.setSpeed(self.movie.speed() + 20)

    def sizeHint(self):
        return self.movie.scaledSize()

    def restartmovie(self):
        """
        Once the animation reaches the end, check if its time to launch, otherwise restart animation
        """
        if self._launchready:
            self.execlaunch()
            return
        self.movie.start()

    def launchwindow(self):
        """
        Save state, defer launch until animation finishes
        """
        self._launchready = True

    def execlaunch(self):
        """
        Launch the mainwindow
        """
        if not self._launching:
            self._launching = True

            app = QApplication.instance()

            from xicam.gui.windows.mainwindow import XicamMainWindow
            self.mainwindow = XicamMainWindow()
            self.timer.stop()

            # Show the QMainWindow
            self.mainwindow.show()
            self.mainwindow.raise_()
            self.mainwindow.activateWindow()
            app.setActiveWindow(self.mainwindow)

            # Stop splashing
            self.hide()
            self.movie.stop()
            self.finish(self.mainwindow)



'''
This patch module adds color gradients like matplotlib's viridis etc. to PyQtGraph
'''
from pyqtgraph.graphicsItems import GradientEditorItem
from collections import OrderedDict
from qtpy.QtWidgets import QTreeWidgetItem

GradientEditorItem.__dict__['Gradients'] = OrderedDict([
    ('thermal', {'ticks': [(0.3333, (185, 0, 0, 255)), (0.6666, (255, 220, 0, 255)), (1, (255, 255, 255, 255)), (0, (0, 0, 0, 255))], 'mode': 'rgb'}),
    ('flame', {'ticks': [(0.2, (7, 0, 220, 255)), (0.5, (236, 0, 134, 255)), (0.8, (246, 246, 0, 255)), (1.0, (255, 255, 255, 255)), (0.0, (0, 0, 0, 255))], 'mode': 'rgb'}),
    ('yellowy', {'ticks': [(0.0, (0, 0, 0, 255)), (0.2328863796753704, (32, 0, 129, 255)), (0.8362738179251941, (255, 255, 0, 255)), (0.5257586450247, (115, 15, 255, 255)), (1.0, (255, 255, 255, 255))], 'mode': 'rgb'} ),
    ('bipolar', {'ticks': [(0.0, (0, 255, 255, 255)), (1.0, (255, 255, 0, 255)), (0.5, (0, 0, 0, 255)), (0.25, (0, 0, 255, 255)), (0.75, (255, 0, 0, 255))], 'mode': 'rgb'}),
    ('viridis', {'ticks': [(0.0, (68, 1, 84, 255)), (0.25, (58, 82, 139, 255)), (0.5, (32, 144, 140, 255)), (0.75, (94, 201, 97, 255)), (1.0, (253, 231, 36, 255))], 'mode': 'rgb'}),
    ('inferno', {'ticks': [(0.0, (0, 0, 3, 255)), (0.25, (87, 15, 109, 255)), (0.5, (187, 55, 84, 255)), (0.75, (249, 142, 8, 255)), (1.0, (252, 254, 164, 255))], 'mode': 'rgb'}),
    ('plasma', {'ticks': [(0.0, (12, 7, 134, 255)), (0.25, (126, 3, 167, 255)), (0.5, (203, 71, 119, 255)), (0.75, (248, 149, 64, 255)), (1.0, (239, 248, 33, 255))], 'mode': 'rgb'}),
    ('magma', {'ticks': [(0.0, (0, 0, 3, 255)), (0.25, (80, 18, 123, 255)), (0.5, (182, 54, 121, 255)), (0.75, (251, 136, 97, 255)), (1.0, (251, 252, 191, 255))], 'mode': 'rgb'}),
    ('spectrum', {'ticks': [(1.0, (255, 0, 255, 255)), (0.0, (255, 0, 0, 255))], 'mode': 'hsv'}),
    ('cyclic', {'ticks': [(0.0, (255, 0, 4, 255)), (1.0, (255, 0, 0, 255))], 'mode': 'hsv'}),
    ('greyclip', {'ticks': [(0.0, (0, 0, 0, 255)), (0.99, (255, 255, 255, 255)), (1.0, (255, 0, 0, 255))], 'mode': 'rgb'}),
    ('grey', {'ticks': [(0.0, (0, 0, 0, 255)), (1.0, (255, 255, 255, 255))], 'mode': 'rgb'}),
])

from pyqtgraph.parametertree import Parameter, ParameterItem, registerParameterType
from pyqtgraph.parametertree.parameterTypes import WidgetParameterItem
from pyqtgraph import ImageView
import numpy as np


class ImageParameterItem(WidgetParameterItem):
    def makeWidget(self):
        self.subItem = QTreeWidgetItem()
        self.addChild(self.subItem)

        w = ImageView()
        w.value = lambda: w.image
        w.setValue = lambda image: w.setImage(np.squeeze(image))
        w.sigChanged = None

        self.hideWidget = False
        return w

    def treeWidgetChanged(self):
        ## TODO: fix so that superclass method can be called
        ## (WidgetParameter should just natively support this style)
        # WidgetParameterItem.treeWidgetChanged(self)
        self.treeWidget().setFirstItemColumnSpanned(self.subItem, True)
        self.treeWidget().setItemWidget(self.subItem, 0, self.widget)

        # for now, these are copied from ParameterItem.treeWidgetChanged
        self.setHidden(not self.param.opts.get('visible', True))
        self.setExpanded(self.param.opts.get('expanded', True))

    def valueChanged(self, param, val, force=False):
        ## called when the parameter's value has changed
        ParameterItem.valueChanged(self, param, val)
        if force or not np.array_equal(val, self.widget.value()):
            self.widget.setValue(val.astype(np.float))
        self.updateDisplayLabel(val)  ## always make sure label is updated, even if values match!

    def updateDefaultBtn(self):
        pass


class ImageParameter(Parameter):
    itemClass = ImageParameterItem

    def __init__(self, *args, **kwargs):
        if 'expanded' not in kwargs: kwargs['expanded'] = False
        super(ImageParameter, self).__init__(*args, **kwargs)

    def setValue(self, value, blockSignal=None):
        """
        Set the value of this Parameter; return the actual value that was set.
        (this may be different from the value that was requested)
        """
        try:
            if blockSignal is not None:
                self.sigValueChanged.disconnect(blockSignal)
            if np.array_equal(self.opts['value'], value):
                return value
            self.opts['value'] = value
            # self.sigValueChanged.emit(self, value)
        finally:
            if blockSignal is not None:
                self.sigValueChanged.connect(blockSignal)

        return value


registerParameterType('ndarray', ImageParameter, override=True)



import pickle
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from xicam.core.execution.workflow import Workflow
from pyqtgraph.parametertree import ParameterTree, Parameter
from xicam.gui.static import path
from xicam.plugins import manager as pluginmanager
from functools import partial


# WorkflowEditor
#  WorkflowProcessEditor
# WorkflowWidget
#  LinearWorkflowView
#   WorkflowModel



class WorkflowEditor(QSplitter):
    sigWorkflowChanged = Signal(object)

    def __init__(self, workflow: Workflow):
        super(WorkflowEditor, self).__init__()
        self.setOrientation(Qt.Vertical)

        self.processeditor = WorkflowProcessEditor()
        self.workflowview = LinearWorkflowView(WorkflowModel(workflow))

        self.addWidget(self.processeditor)
        self.addWidget(WorkflowWidget(self.workflowview))

        self.workflowview.sigShowParameter.connect(
            lambda parameter: self.processeditor.setParameters(parameter, showTop=False))

        workflow.attach(partial(self.sigWorkflowChanged.emit, workflow))


class WorkflowProcessEditor(ParameterTree):
    pass


class WorkflowWidget(QWidget):
    sigAddFunction = Signal(object)

    def __init__(self, workflowview: QAbstractItemView):
        super(WorkflowWidget, self).__init__()

        self.view = workflowview

        self.toolbar = QToolBar()
        addfunctionmenu = QToolButton()
        functionmenu = QMenu()
        for plugin in pluginmanager.getPluginsOfCategory('ProcessingPlugin'):
            functionmenu.addAction(plugin.name, partial(self.addProcess, plugin.plugin_object,
                                                        autoconnectall=True))
        addfunctionmenu.setMenu(functionmenu)
        addfunctionmenu.setIcon(QIcon(path('icons/addfunction.png')))
        addfunctionmenu.setText('Add Function')
        addfunctionmenu.setPopupMode(QToolButton.InstantPopup)
        self.toolbar.addWidget(addfunctionmenu)
        # self.toolbar.addAction(QIcon(path('icons/up.png')), 'Move Up')
        # self.toolbar.addAction(QIcon(path('icons/down.png')), 'Move Down')
        self.toolbar.addAction(QIcon(path('icons/folder.png')), 'Load Workflow')
        self.toolbar.addAction(QIcon(path('icons/trash.png')), 'Clear Workflow')

        v = QVBoxLayout()
        v.addWidget(self.view)
        v.addWidget(self.toolbar)
        v.setContentsMargins(0, 0, 0, 0)
        self.setLayout(v)

    def addProcess(self, process, autoconnectall=True):
        self.view.model().workflow.addProcess(process(), autoconnectall)
        print('selected new row:', self.view.model().rowCount() - 1)
        self.view.setCurrentIndex(self.view.model().index(self.view.model().rowCount() - 1, 0))

class LinearWorkflowView(QTableView):
    sigShowParameter = Signal(object)

    def __init__(self, workflowmodel=None, *args, **kwargs):
        super(LinearWorkflowView, self).__init__(*args, **kwargs)
        self.setItemDelegateForColumn(0, DisableDelegate(self))
        self.setItemDelegateForColumn(2, DeleteDelegate(self))

        self.setModel(workflowmodel)
        workflowmodel.workflow.attach(self.selectionChanged)

        self.horizontalHeader().close()
        # self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setResizeMode(QHeaderView.ResizeToContents)
        self.horizontalHeader().setResizeMode(1, QHeaderView.Stretch)

        # self.horizontalHeader().setSectionMovable(True)
        # self.horizontalHeader().setDragEnabled(True)
        # self.horizontalHeader().setDragDropMode(QAbstractItemView.InternalMove)
        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDragEnabled(True)
        self.setDropIndicatorShown(True)
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

    def selectionChanged(self, selected=None, deselected=None):
        if self.selectedIndexes() and self.selectedIndexes()[0].row() < self.model().rowCount():
            process = self.model().workflow._processes[self.selectedIndexes()[0].row()]
            self.sigShowParameter.emit(process.parameter)
        else:
            self.sigShowParameter.emit(Parameter(name='empty'))
        for child in self.children():
            if hasattr(child, 'repaint'): child.repaint()


class WorkflowModel(QAbstractTableModel):
    def __init__(self, workflow: Workflow):
        self.workflow = workflow
        super(WorkflowModel, self).__init__()


        self.workflow.attach(partial(self.layoutChanged.emit))

    def mimeTypes(self):
        return ['text/plain']

    def mimeData(self, indexes):
        mimedata = QMimeData()
        mimedata.setText(str(indexes[0].row()))
        return mimedata

    def dropMimeData(self, data, action, row, column, parent):
        srcindex = int(data.text())
        process = self.workflow._processes[srcindex]
        self.workflow.removeProcess(process)
        self.workflow.insertProcess(parent.row(), process)
        self.workflow.autoConnectAll()
        return True

    def supportedDropActions(self):
        return Qt.CopyAction | Qt.MoveAction

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemIsEnabled
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable | \
               Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled

    def rowCount(self, *args, **kwargs):
        return len(self.workflow._processes)

    def columnCount(self, *args, **kwargs):
        return 3

    def data(self, index, role):
        process = self.workflow._processes[index.row()]
        if not index.isValid():
            return None
        elif role != Qt.DisplayRole:
            return None
        elif index.column() == 0:
            return partial(self.workflow.toggleDisableProcess, process, autoconnectall=True)
        elif index.column() == 1:
            return getattr(process, 'name', process.__class__.__name__)
        elif index.column() == 2:
            return partial(self.workflow.removeProcess, index=index.row(), autoconnectall=True)
        return ''

    def headerData(self, col, orientation, role):
        return None


class DeleteDelegate(QItemDelegate):
    def __init__(self, parent):
        super(DeleteDelegate, self).__init__(parent)
        self._parent = parent

    def paint(self, painter, option, index):
        if not self._parent.indexWidget(index):
            button = QToolButton(self.parent(), )
            button.setAutoRaise(True)
            button.setText('Delete Operation')
            button.setIcon(QIcon(path('icons/trash.png')))
            sp = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
            sp.setWidthForHeight(True)
            button.setSizePolicy(sp)
            button.clicked.connect(index.data())

            self._parent.setIndexWidget(index, button)


class DisableDelegate(QItemDelegate):
    def __init__(self, parent):
        super(DisableDelegate, self).__init__(parent)
        self._parent = parent

    def paint(self, painter, option, index):
        if not self._parent.indexWidget(index):
            button = QToolButton(self.parent())
            button.setText('i')
            button.setAutoRaise(True)
            icon = QIcon()
            icon.addPixmap(QPixmap(path('icons/enable.png')), state=icon.Off)
            icon.addPixmap(QPixmap(path('icons/disable.png')), state=icon.On)
            button.setIcon(icon)
            button.setCheckable(True)
            sp = QSizePolicy()
            sp.setWidthForHeight(True)
            button.setSizePolicy(sp)
            button.clicked.connect(index.data())

            self._parent.setIndexWidget(index, button)



from pyqtgraph import ImageView
import numpy as np


class DynImageView(ImageView):
    def __init__(self, *args, **kwargs):
        super(DynImageView, self).__init__(*args, **kwargs)

        # Use Viridis by default
        self.setPredefinedGradient('viridis')

        # Shrink LUT
        self.getHistogramWidget().setMinimumWidth(10)

        # Don't invert Y axis
        self.view.invertY(False)

    def quickMinMax(self, data):
        """
        Estimate the min/max values of *data* by subsampling. MODIFIED TO USE THE 99TH PERCENTILE instead of max.
        """
        if data.size > 1e6:
            data = data[len(data) // 2]
        return np.nanmin(data), np.nanpercentile(data, 99)

    def setImage(self, img, autoRange=True, autoLevels=True, levels=None, axes=None, xvals=None, pos=None, scale=None,
                 transform=None, autoHistogramRange=True):
        super(DynImageView, self).setImage(img, autoRange, autoLevels, levels, axes, xvals, pos, scale, transform,
                                           autoHistogramRange)
        if len(img.shape) > 2 and img.shape[0] == 1:
            self.ui.roiPlot.hide()



from functools import partial

from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from xicam.plugins.GUIPlugin import PanelState
from yapsy import PluginInfo

from xicam.plugins import manager as pluginmanager
from xicam.plugins import observers as pluginobservers
from xicam.core import msg
from ..widgets import defaultstage
from .settings import ConfigDialog
from ..static import path


class XicamMainWindow(QMainWindow):
    """
    The Xi-cam main window. Includes layout for various panels and mechanism to position GUIPlugin contents into panels.
    """

    def __init__(self):
        super(XicamMainWindow, self).__init__()

        # Set icon
        self.setWindowIcon(QIcon(QPixmap(str(path('icons/xicam.gif')))))

        # Set size and position
        self.setGeometry(0, 0, 1000, 600)
        frameGm = self.frameGeometry()
        screen = QApplication.desktop().screenNumber(QApplication.desktop().cursor().pos())
        centerPoint = QApplication.desktop().screenGeometry(screen).center()
        frameGm.moveCenter(centerPoint)
        self.move(frameGm.topLeft())

        # Init child widgets to None
        self.topwidget = self.leftwidget = self.rightwidget = self.bottomwidget = self.lefttopwidget = \
            self.righttopwidget = self.leftbottomwidget = self.rightbottomwidget = None

        # Setup appearance
        self.setWindowTitle('Xi-cam')

        # Restore Settings
        ConfigDialog().restore()

        # Load plugins
        pluginmanager.collectPlugins()

        # Setup center/toolbar/statusbar/progressbar
        self.pluginmodewidget = pluginModeWidget()
        self.pluginmodewidget.sigSetStage.connect(self.setStage)
        self.pluginmodewidget.sigSetGUIPlugin.connect(self.setGUIPlugin)
        self.addToolBar(self.pluginmodewidget)
        self.setStatusBar(QStatusBar(self))
        msg.progressbar = QProgressBar(self)
        msg.progressbar.hide()
        msg.statusbar = self.statusBar()
        self.statusBar().addPermanentWidget(msg.progressbar)
        self.setCentralWidget(QStackedWidget())
        # NOTE: CentralWidgets are force-deleted when replaced, even if the object is still referenced;
        # To avoid this, a QStackedWidget is used for the central widget.

        # Setup menubar
        menubar = self.menuBar()
        file = QMenu('&File', parent=menubar)
        menubar.addMenu(file)
        file.addAction('Se&ttings', self.showSettings, shortcut=QKeySequence(Qt.CTRL + Qt.ALT + Qt.Key_S))
        file.addAction('E&xit', self.close)
        help = QMenu('&Help', parent=menubar)
        menubar.addMenu(help)

        # Initialize layout with first plugin
        self._currentGUIPlugin = None
        self.build_layout()
        if pluginmanager.getPluginsOfCategory("GUIPlugin"):
            # self._currentGUIPlugin = pluginmanager.getPluginsOfCategory("GUIPlugin")[0]
            self.populate_layout()

        # Make F key bindings
        fkeys = [Qt.Key_F1, Qt.Key_F2, Qt.Key_F3, Qt.Key_F4, Qt.Key_F5, Qt.Key_F6,
                 Qt.Key_F7, Qt.Key_F8, Qt.Key_F9, Qt.Key_F10, Qt.Key_F11, Qt.Key_F12]
        self.Fshortcuts = [QShortcut(QKeySequence(key), self) for key in fkeys]
        for i in range(12):
            self.Fshortcuts[i].activated.connect(partial(self.setStage, i))

        # Wireup default widgets
        defaultstage['left'].sigOpen.connect(self.open)
        defaultstage['left'].sigOpen.connect(print)
        defaultstage['left'].sigPreview.connect(defaultstage['lefttop'].preview_header)

    def open(self, header):
        self.currentGUIPlugin.plugin_object.appendHeader(header)

    def showSettings(self):
        self._configdialog = ConfigDialog()
        self._configdialog.show()

    @Slot(int)
    def setStage(self, i: int):
        """
        Set the current Stage/Layout/Plugin mode to number i in its sequence. Triggered by menu (TODO) or F keybindings.

        Parameters
        ----------
        i   : int
        """
        plugin = self.currentGUIPlugin.plugin_object
        if i < len(plugin.stages):
            plugin.stage = list(plugin.stages.values())[i]
            self.populate_layout()

    @Slot(int)
    def setGUIPlugin(self, i: int):
        self.currentGUIPlugin = pluginmanager.getPluginsOfCategory('GUIPlugin')[i]

    @property
    def currentGUIPlugin(self) -> PluginInfo:
        return self._currentGUIPlugin

    @currentGUIPlugin.setter
    def currentGUIPlugin(self, plugininfo: PluginInfo):
        if plugininfo != self._currentGUIPlugin:
            self._currentGUIPlugin = plugininfo
            self.populate_layout()

    def build_layout(self):
        # TODO: Allow save/restore https://stackoverflow.com/questions/14288635/any-easy-way-to-store-dock-widows-layout-and-sizes-in-settings-with-qt

        # Clear out docks
        for dockwidget in self.findChildren(QDockWidget):
            if dockwidget.parent() == self:
                self.removeDockWidget(dockwidget)

        # Make new docks
        self.topwidget = QDockWidget(parent=self)
        self.leftwidget = QDockWidget(parent=self)
        self.rightwidget = QDockWidget(parent=self)
        self.bottomwidget = QDockWidget(parent=self)
        self.lefttopwidget = QDockWidget(parent=self)
        self.righttopwidget = QDockWidget(parent=self)
        self.leftbottomwidget = QDockWidget(parent=self)
        self.rightbottomwidget = QDockWidget(parent=self)

        # Place the docks
        self.addDockWidget(Qt.LeftDockWidgetArea, self.lefttopwidget)
        self.addDockWidget(Qt.RightDockWidgetArea, self.righttopwidget)
        self.addDockWidget(Qt.TopDockWidgetArea, self.topwidget)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.leftwidget)
        self.addDockWidget(Qt.RightDockWidgetArea, self.rightwidget)
        self.addDockWidget(Qt.BottomDockWidgetArea, self.bottomwidget)
        self.addDockWidget(Qt.LeftDockWidgetArea, self.leftbottomwidget)
        self.addDockWidget(Qt.RightDockWidgetArea, self.rightbottomwidget)

        # Adjust spanning
        self.setCorner(Qt.TopLeftCorner, Qt.LeftDockWidgetArea)
        self.setCorner(Qt.TopRightCorner, Qt.RightDockWidgetArea)
        self.setCorner(Qt.BottomLeftCorner, Qt.LeftDockWidgetArea)
        self.setCorner(Qt.BottomRightCorner, Qt.RightDockWidgetArea)

    def populate_layout(self):
        # Get current stage
        if self.currentGUIPlugin:
            stage = self.currentGUIPlugin.plugin_object.stage
        else:
            stage = defaultstage

        # Set center contents
        self.centralWidget().addWidget(stage.centerwidget)
        self.centralWidget().setCurrentWidget(stage.centerwidget)

        # Set visibility based on panel state and (TODO) insert default widgets when defaulted
        for position in ['top','left','right','bottom','lefttop','righttop','leftbottom','rightbottom']:
            self.populate_hidden(stage, position)
            self.populate_position(stage, position)

    def populate_hidden(self, stage, position):
        getattr(self,position+'widget').setHidden((stage[position] == PanelState.Disabled) or
                                                  (stage[position] == PanelState.Defaulted and
                                                   defaultstage[position] == PanelState.Defaulted))

    def populate_position(self, stage, position:str):
        if isinstance(stage[position], QWidget):
            getattr(self,position+'widget').setWidget(stage[position])
        elif stage[position] == PanelState.Defaulted:
            if not defaultstage[position]==PanelState.Defaulted:
                getattr(self,position+'widget').setWidget(defaultstage[position])
        elif isinstance(stage[position], type):
            raise TypeError(
                f'A type is not acceptable value for stages. You must instance this class: {stage[position]}, {position}')

    def mousePressEvent(self, event):
        focused_widget = QApplication.focusWidget()
        if isinstance(focused_widget, QSpinBox):
            focused_widget.clearFocus()
        super(XicamMainWindow, self).mousePressEvent(event)


class pluginModeWidget(QToolBar):
    """
    A series of styled QPushButtons with pipe characters between them. Used to switch between plugin modes.
    """
    sigSetStage = Signal(int)
    sigSetGUIPlugin = Signal(int)

    def __init__(self):
        super(pluginModeWidget, self).__init__()

        self.GUIPluginActionGroup = QActionGroup(self)

        # Setup font
        self.font = QFont('Zero Threes')
        self.font.setPointSize(16)

        # Align right
        self.setLayoutDirection(Qt.RightToLeft)

        # Subscribe to plugins
        pluginobservers.append(self)

        # Build children
        self.pluginsChanged()

    def pluginsChanged(self):
        self.showGUIPlugins()

    def fadeOut(self, callback, distance=-20):
        duration = 200
        self._effects = []
        for action in self.actions():
            for widget in action.associatedWidgets():
                if widget is not self:
                    a = QPropertyAnimation(widget, b'pos', widget)
                    a.setStartValue(widget.pos())
                    a.setEndValue(widget.pos() + QPoint(0, distance))
                    self._effects.append(a)
                    a.setDuration(duration)
                    a.setEasingCurve(QEasingCurve.OutBack)
                    a.start(QPropertyAnimation.DeleteWhenStopped)

                    effect = QGraphicsOpacityEffect(self)
                    widget.setGraphicsEffect(effect)
                    self._effects.append(effect)
                    b = QPropertyAnimation(effect, b'opacity')
                    self._effects.append(b)
                    b.setDuration(duration)
                    b.setStartValue(1)
                    b.setEndValue(0)
                    b.setEasingCurve(QEasingCurve.OutBack)
                    b.start(QPropertyAnimation.DeleteWhenStopped)
                    b.finished.connect(partial(self.removeAction, action))
                    b.finished.connect(callback)
        if not self.actions(): callback()

    def fadeIn(self):
        self._effects = []
        for action in self.actions():
            effect = QGraphicsOpacityEffect(self)
            self._effects.append(effect)
            for widget in action.associatedWidgets():
                if widget is not self: widget.setGraphicsEffect(effect)
            a = QPropertyAnimation(effect, b'opacity')
            self._effects.append(a)
            a.setDuration(1000)
            a.setStartValue(0)
            a.setEndValue(1)
            a.setEasingCurve(QEasingCurve.OutBack)
            a.start(QPropertyAnimation.DeleteWhenStopped)

    def showStages(self, plugin):
        self.sigSetGUIPlugin.emit(plugin)
        if len(self.parent().currentGUIPlugin.plugin_object.stages) > 1:
            names = self.parent().currentGUIPlugin.plugin_object.stages.keys()
            self.fadeOut(callback=partial(self.mkButtons, names=names, callback=self.sigSetStage.emit,
                                          parent=self.parent().currentGUIPlugin.name))

    def showGUIPlugins(self):
        plugins = pluginmanager.getPluginsOfCategory('GUIPlugin')
        # TODO: test deactivated plugins
        names = [plugin.name for plugin in plugins if hasattr(plugin, 'is_activated') and (plugin.is_activated or True)]
        self.fadeOut(callback=partial(self.mkButtons, names=names, callback=self.showStages), distance=20)

    def mkButtons(self, names, callback, parent=None):
        # Remove+delete previous children
        layout = self.layout()
        for i in reversed(range(layout.count())):
            layout.itemAt(i).widget().setParent(None)

        if parent:
            action = QAction('↑', self)
            action.setFont(self.font)
            action.triggered.connect(self.showGUIPlugins)
            action.setProperty('isMode', True)
            self.addAction(action)
            # Make separator pipe
            label = QAction('|', self)
            label.setFont(self.font)
            label.setDisabled(True)
            self.addAction(label)

        # Loop over each "GUIPlugin" plugin
        for i, name in zip(reversed(range(len(names))), reversed(list(names))):
            action = QAction(name, self)
            action.triggered.connect(partial(callback, i))
            action.setFont(self.font)
            action.setProperty('isMode', True)
            action.setCheckable(True)
            action.setActionGroup(self.GUIPluginActionGroup)
            self.addAction(action)

            # Make separator pipe
            label = QAction('|', self)
            label.setFont(self.font)
            label.setDisabled(True)
            self.addAction(label)


        # Remove last separator
        if self.layout().count():
            self.layout().takeAt(self.layout().count() - 1).widget().deleteLater()  # Delete the last pipe symbol

        if parent:
            # Make separator pipe
            label = QAction('>', self)
            label.setFont(self.font)
            label.setDisabled(True)
            self.addAction(label)

            action = QAction(parent, self)
            action.setFont(self.font)
            action.setProperty('isMode', True)
            action.setDisabled(True)
            action.setActionGroup(self.GUIPluginActionGroup)
            self.addAction(action)

        self.fadeIn()



from xicam.plugins.GUIPlugin import GUILayout
from .motd import MOTD
from .dataresourcebrowser import DataResourceBrowser
from .previewwidget import PreviewWidget

from qtpy.QtWidgets import QApplication

if QApplication.instance():
    defaultstage = GUILayout(center=MOTD(),
                         left=DataResourceBrowser(),
                         lefttop=PreviewWidget())


from qtpy.QtGui import *
from qtpy.QtWidgets import *
from xicam.gui.static import path

from xicam.plugins import SettingsPlugin


class VenvsSettingsPlugin(SettingsPlugin):
    name = 'Virtual Environments'

    def __init__(self):
        self.widget = QLabel('test')
        super(VenvsSettingsPlugin, self).__init__(QIcon(str(path('icons/python.png'))),
                                                  self.name,
                                                  self.widget)

    def save(self):
        return None  # self.parameter.saveState()

    def restore(self, state):
        pass  # self.parameter.restoreState(state)



from qtpy.QtCore import QSize
from qtpy.QtGui import QFont
from qtpy.QtWidgets import QSizePolicy
from pyqtgraph import ImageItem, TextItem, GraphicsLayoutWidget
import os
from xicam.core.data import NonDBHeader
import numpy as np


class PreviewWidget(GraphicsLayoutWidget):
    def __init__(self):
        super(PreviewWidget, self).__init__()
        self.setMinimumHeight(250)
        self.setMinimumWidth(250)
        self.view = self.addViewBox(lockAspect=True, enableMenu=False)
        self.imageitem = ImageItem()
        self.textitem = TextItem(anchor=(0.5, 0))
        self.textitem.setFont(QFont('Zero Threes'))
        self.imgdata = None

        self.view.addItem(self.imageitem)
        self.view.addItem(self.textitem)
        self.textitem.hide()
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        # def textItemBounds(axis, frac=1.0, orthoRange=None):
        #     b = self.textitem.boundingRect()
        #     sx, sy = self.view.viewPixelSize()
        #     x, y = sx*b.width(), sy*b.height()
        #     if axis == 0: return (-x/2, x/2)
        #     if axis == 1: return (0, y)
        #
        # self.textitem.dataBounds = textItemBounds

    def sizeHint(self):
        return QSize(250, 250)

    def preview_header(self, header: NonDBHeader):
        try:
            data = header.meta_array()[0]
            self.setImage(data)
        except IndexError:
            self.imageitem.clear()
            self.setText('UNKNOWN DATA FORMAT')

    def setImage(self, imgdata):
        self.imageitem.clear()
        self.textitem.hide()
        self.imgdata = imgdata
        self.imageitem.setImage(np.rot90(np.log(self.imgdata * (self.imgdata > 0) + (self.imgdata < 1)), 3),
                                autoLevels=True)
        self.view.autoRange()

    def setText(self, text):
        self.textitem.setText(text)
        self.imageitem.clear()
        self.textitem.setVisible(True)
        self.view.autoRange()
